@@ -1,15 +1,18 @@
 from abc import ABC, abstractmethod
 from collections.abc import Iterable
 
-from babel.numbers import format_number
+from babel.numbers import format_decimal
+from flask import url_for
 from flask_babel import get_locale
+from werkzeug.datastructures import MultiDict
 from whoosh.fields import ID
 from whoosh.query import Prefix, Term
 
-from . import extractors, renderers
-from .codecs import BaseFacetCodec, CollectionFacetCodec, IdentityFieldCodec
-from .text import slugify, sort_normalize
-from .tree import Tree
+from kerko import extractors, renderers
+from kerko.codecs import (BaseFacetCodec, CollectionFacetCodec,
+                          IdentityFieldCodec)
+from kerko.text import slugify, sort_normalize
+from kerko.tree import Tree
 
 
 class ScopeSpec:
@@ -26,6 +29,37 @@ class ScopeSpec:
         self.weight = weight
         self.help_text = help_text
 
+    def add_keywords(self, value, active_keywords=None):
+        """
+        Add a value for this scope to active keywords.
+
+        :param string value: The value to add.
+
+        :param MultiDict active_keywords: The active keywords to derive from.
+
+        :return MultiDict: A copy of the keywords with the added value.
+        """
+        new_keywords = active_keywords.deepcopy() if active_keywords else MultiDict()
+        new_keywords.add(self.key, value.strip())
+        return new_keywords
+
+    def remove_keywords(self, value, active_keywords):
+        """
+        Remove a value for this facet from active keywords.
+
+        :param string value: The value to remove.
+
+        :param MultiDict active_keywords: The active keywords to derive from.
+
+        :return MultiDict: A copy of the keywords with the value removed.
+        """
+        new_keywords = active_keywords.deepcopy()
+        if new_keywords:
+            new_values = [v for v in new_keywords.poplist(self.key) if v != value]
+            if new_values:
+                new_keywords.setlist(self.key, new_values)
+        return new_keywords
+
 
 class BaseFieldSpec(ABC):
 
@@ -53,9 +87,9 @@ class BaseFieldSpec(ABC):
         self.field_type = field_type
         self.extractor = extractor
 
-    def extract_to_document(self, document, item_context, library_context):
+    def extract_to_document(self, document, item, library_context):
         """Extract the value of this element from a Zotero item."""
-        self.extractor.extract_and_store(document, item_context, library_context, self)
+        self.extractor.extract_and_store(document, item, library_context, self)
 
 
 class FieldSpec(BaseFieldSpec):
@@ -95,6 +129,8 @@ class FacetSpec(BaseFieldSpec):
             title,
             filter_key,
             weight=0,
+            initial_limit=0,
+            initial_limit_leeway=0,
             codec=None,
             missing_label=None,
             sort_key=None,
@@ -102,7 +138,6 @@ class FacetSpec(BaseFieldSpec):
             item_view=True,
             allow_overlap=True,
             query_class=None,
-            collapsible=True,
             renderer=None,
             **kwargs
     ):
@@ -111,25 +146,38 @@ class FacetSpec(BaseFieldSpec):
 
         :param str title: Title of the facet.
 
-        :param str filter_key: Key to use in URLs when filtering with this facet.
+        :param str filter_key: Key to use in URLs when filtering with this
+            facet.
 
         :param int weight: Determine the position of this facet relative to the
             others.
 
+        :param int initial_limit: Maximum number of filters to show by default
+            under this facet. Excess filters will be shown if the user clicks a
+            "view more" button. Defaults to 0 (no limit).
+
+        :param int initial_limit_leeway: If the number of filters under this
+            facet exceeds `initial_limit` by this tolerance margin or less, all
+            filters will be shown. Defaults to 0 (no tolerance margin).
+
+        :param str missing_label: Label to use for items that do not have any
+            value for this facet. Defaults to `None` (show no label at all).
+
         :param BaseFacetCodec codec: Value encoder/decoder for this facet.
 
         :param bool item_view: Show this facet on item view pages.
 
         :param `renderers.Renderer` renderer: A renderer for this facet. The
             rendering context provides the following variables:
+
             - `spec`: The `FacetSpec` instance.
             - `items`: The facet values retrieved by the search query.
             - `mode`: A string whose value is one of the following:
                 - `'search'`: Displaying the facet on a search results page.
-                - `'field'`: Displaying the facet on a full bibliographic
-                  record page.
-                - `'breadbox'`: Displaying the facet as a search criteria in
-                  the breadbox.
+                - `'field'`: Displaying the facet on a full bibliographic record
+                  page.
+                - `'breadbox'`: Displaying the facet as a search criteria in the
+                  breadbox.
 
         .. seealso: Additional :meth:`BaseFieldSpec.__init__` arguments.
         """
@@ -137,6 +185,8 @@ class FacetSpec(BaseFieldSpec):
         self.title = title
         self.filter_key = filter_key
         self.weight = weight
+        self.initial_limit = initial_limit
+        self.initial_limit_leeway = initial_limit_leeway
         self.codec = codec or BaseFacetCodec()
         self.missing_label = missing_label
         self.sort_key = sort_key
@@ -144,7 +194,6 @@ class FacetSpec(BaseFieldSpec):
         self.item_view = item_view
         self.allow_overlap = allow_overlap
         self.query_class = query_class or Term
-        self.collapsible = collapsible
         self.renderer = renderer or renderers.TemplateResolverRenderer(
             'kerko/_facet_{mode}.html.jinja2'
         )
@@ -160,25 +209,25 @@ class FacetSpec(BaseFieldSpec):
     @abstractmethod
     def add_filter(self, value, active_filters):
         """
-        Build a query string that adds a value for this facet to the active filters.
+        Add a value for this facet to active filters.
 
-        :param string value: The value to add to the active filters.
+        :param string value: The value to add.
 
         :param MultiDict active_filters: The active filters to derive from.
 
-        :return MultiDict: The resulting filters.
+        :return MultiDict: A copy of the filters with the added value.
         """
 
     @abstractmethod
     def remove_filter(self, value, active_filters):
         """
-        Build a query string that removes a value for this facet from the active filters.
+        Remove a value for this facet from active filters.
 
-        :param string value: The value to remove from the active filters.
+        :param string value: The value to remove.
 
         :param MultiDict active_filters: The active filters to derive from.
 
-        :return MultiDict: The resulting filters.
+        :return MultiDict: A copy of the filters with the value removed.
         """
 
     @abstractmethod
@@ -199,21 +248,20 @@ class FacetSpec(BaseFieldSpec):
     def sort_items(self, items):
         if self.sort_key is None:
             return
-        # Sort items with multiple-keys, using tuples. The first tuple element
-        # (the primary sort key), gets inserted before the other sort keys. It
-        # is a boolean value that ensures that empty labels (those of missing
-        # values) appear last regardless of the sort direction.
-        #
-        # Note: Count values are multiplied by -1 to reverse their order.
-        # Because there is only one 'reverse' flag for all sort keys, the
-        # reverse order for counts is, counterintuitively, ascending. When
-        # ordering facet values by count, we usually want descending order, so
-        # if the reverse flag False in that case, it is possible to get the
-        # second sort key, usually 'label' sorted in alphabetical order.
+
+        # Sort items based on multiple-keys.
         return sorted(
             items,
             key=lambda x: (
-                bool(x['label']) if self.sort_reverse else not x['label'], *[
+                # First sort key: show active items first.
+                not x['remove_url'],
+                # Second sort key: show items with missing labels last.
+                bool(x['label']) if self.sort_reverse else not x['label'],
+                # Third sort key: sort items according to the facet's specified
+                # sort keys. If 'count' is used as key, multiply the count value
+                # by -1 to reverse order (because the desired default when
+                # ordering by count is the descending order).
+                *[
                     x[k] * -1 if k == 'count' else
                     (sort_normalize(x[k]) if isinstance(x[k], str) else x[k])
                     for k in self.sort_key
@@ -251,19 +299,47 @@ class FlatFacetSpec(FacetSpec):
 
     def build(self, results, criteria, active_only=False):
         items = []
-        for value, count in results:
+        for value, count in results.items():
             if value or self.missing_label:
                 value, label = self.decode(value, default_value=value, default_label=value)
-                remove_url = criteria.build_remove_filter_url(self, value)
+                new_filters = self.remove_filter(value, criteria.filters)
+                if new_filters:
+                    remove_url = url_for(
+                        '.search',
+                        **criteria.params(
+                            filters=new_filters,
+                            options={
+                                'page': None,
+                                'page-len': None,
+                                'id': None,
+                            },
+                        )
+                    )
+                else:
+                    remove_url = None
                 if remove_url or active_only:
                     add_url = None
                 else:
-                    add_url = criteria.build_add_filter_url(self, value)
+                    new_filters = self.add_filter(value, criteria.filters)
+                    if new_filters:
+                        add_url = url_for(
+                            '.search',
+                            **criteria.params(
+                                filters=new_filters,
+                                options={
+                                    'page': None,
+                                    'page-len': None,
+                                    'id': None,
+                                },
+                            )
+                        )
+                    else:
+                        add_url = None
                 if remove_url or add_url:  # Only items with an URL get displayed.
                     items.append({
                         'label': label,
                         'count': count,
-                        'count_formatted': format_number(count, locale=get_locale()),
+                        'count_formatted': format_decimal(count, locale=get_locale()),
                         'remove_url': remove_url,
                         'add_url': add_url,
                     })
@@ -356,19 +432,44 @@ class TreeFacetSpec(FacetSpec):
 
     def build(self, results, criteria, active_only=False):
         tree = Tree()
-        for value, count in results:
+        for value, count in results.items():
             if value or self.missing_label:
                 value, label = self.decode(value, default_value=value, default_label=value)
-                remove_url = criteria.build_remove_filter_url(self, value)
+                new_filters = self.remove_filter(value, criteria.filters)
+                if new_filters:
+                    remove_url = url_for(
+                        '.search',
+                        **criteria.params(
+                            filters=new_filters,
+                            options={
+                                'page': None,
+                                'page-len': None,
+                                'id': None,
+                            },
+                        )
+                    )
+                else:
+                    remove_url = None
                 if remove_url or active_only:
                     add_url = None
                 else:
-                    add_url = criteria.build_add_filter_url(self, value)
-                if remove_url or add_url:  # Only items with an URL get displayed.
-                    if value:
-                        path = value.split(sep=self.path_separator)
+                    new_filters = self.add_filter(value, criteria.filters)
+                    if new_filters:
+                        add_url = url_for(
+                            '.search',
+                            **criteria.params(
+                                filters=new_filters,
+                                options={
+                                    'page': None,
+                                    'page-len': None,
+                                    'id': None,
+                                },
+                            )
+                        )
                     else:
-                        path = [value]
+                        add_url = None
+                if remove_url or add_url:  # Only items with an URL get displayed.
+                    path = value.split(sep=self.path_separator)
 
                     # Build the tree path. Part of the path may or may not already
                     # exist as the facet values are not ordered.
@@ -377,9 +478,10 @@ class TreeFacetSpec(FacetSpec):
                         node = node['children'][component]
                     # Add data at the leaf.
                     node['node'] = {
+                        'id': '-'.join(path),
                         'label': label,
                         'count': count,
-                        'count_formatted': format_number(count, locale=get_locale()),
+                        'count_formatted': format_decimal(count, locale=get_locale()),
                         'remove_url': remove_url,
                         'add_url': add_url,
                     }
@@ -472,7 +574,7 @@ class SortSpec:
 
 class CitationFormatSpec:
     """
-    Specifies a citation download format.
+    Specifies a record download format.
 
     This is a configuration element, with no effect on the search index schema.
     """
@@ -490,21 +592,20 @@ class CitationFormatSpec:
             group_item_delimiter=''
     ):
         """
-        Initialize a citation download format.
+        Initialize a format.
 
-        :param str key: Key of this citation format.
+        :param str key: Key of this format.
 
-        :param FieldSpec field: `FieldSpec` instance associated to this citation
-            format.
+        :param FieldSpec field: `FieldSpec` instance associated to this format.
 
-        :param str label: Label of this citation format.
+        :param str label: Label of this format.
 
-        :param int weight: Determine the position of this citation format
-            relative to the others in lists.
+        :param int weight: Determine the position of this format relative to the
+            others in lists.
 
-        :param str extension: File extension of this citation format.
+        :param str extension: File extension of this format.
 
-        :param str mime_type: MIME type of this citation format.
+        :param str mime_type: MIME type of this format.
 
         :param str group_format: Format string for wrapping multiple entries.
 
